<?php
declare(strict_types=1);

namespace FiloBlu\Refilo\Remote\Entity\Provider;

use FiloBlu\Refilo\Model\CrosssellFactory;
use FiloBlu\Refilo\Model\RelatedFactory;
use FiloBlu\Refilo\Model\UpsellFactory;
use FiloBlu\Refilo\Remote\Entity\EntityProviderInterface;
use FiloBlu\Refilo\Remote\Entity\ProductRelated;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\ResourceModel\Product\Collection;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\Framework\App\Area;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Api\Data\WebsiteInterface;
use Magento\Store\Model\App\Emulation;
use Magento\Store\Model\StoreManagerInterface;
use Magento\TargetRule\Helper\Data;
use Magento\TargetRule\Model\Rule;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 * Class RelatedProvider
 * @package FiloBlu\Refilo\Remote\Entity\Provider
 */
class RelatedProvider extends BaseProvider
{
    /**
     * @var UpsellFactory
     */
    private $upsellFactory;

    /**
     * @var CollectionFactory
     */
    private $productCollectionFactory;

    /**
     * @var RelatedFactory
     */
    private $relatedFactory;

    /**
     * @var Data
     */
    private $targetHelper;

    /**
     * @var Emulation
     */
    private $emulation;

    /**
     * @var LoggerInterface
     */
    private $logger;

    /**
     * @var StoreManagerInterface
     */
    private $storeManager;

    /**
     * @var int
     */
    private $upsellBehaviour;

    /**
     * @var int
     */
    private $relatedBehaviour;

    /**
     * @var int
     */
    private $crosssellBehaviour;

    /**
     * @var CrosssellFactory
     */
    private $crosssellFactory;

    /**
     * RelatedProvider constructor.
     * @param CollectionFactory $productCollectionFactory
     * @param UpsellFactory $upsellFactory
     * @param RelatedFactory $relatedFactory
     * @param Emulation $emulation
     * @param Data $targetHelper
     * @param StoreManagerInterface $storeManager
     * @param ScopeConfigInterface $scopeConfig
     * @param LoggerInterface $logger
     */
    public function __construct(
        CollectionFactory     $productCollectionFactory,
        UpsellFactory         $upsellFactory,
        RelatedFactory        $relatedFactory,
        Emulation             $emulation,
        Data                  $targetHelper,
        StoreManagerInterface $storeManager,
        ScopeConfigInterface  $scopeConfig,
        LoggerInterface       $logger,
        CrosssellFactory      $crosssellFactory
    )
    {
        parent::__construct($scopeConfig);
        $this->upsellFactory = $upsellFactory;
        $this->productCollectionFactory = $productCollectionFactory;
        $this->relatedFactory = $relatedFactory;
        $this->targetHelper = $targetHelper;
        $this->emulation = $emulation;
        $this->logger = $logger;
        $this->storeManager = $storeManager;
        $this->crosssellFactory = $crosssellFactory;
    }

    /**
     * @param StoreInterface $store
     * @return EntityProviderInterface
     */
    public function forStore(StoreInterface $store): EntityProviderInterface
    {
        return $this;
    }

    /**
     * @param WebsiteInterface $website
     * @return EntityProviderInterface
     */
    public function forWebsite(WebsiteInterface $website): EntityProviderInterface
    {
        return $this;
    }

    /**
     *
     */
    public function rewind()
    {
        $this->items = [];

        /** @var Collection $productCollection */
        $productCollection = $this->productCollectionFactory->create();
        if ($this->ids) {
            $productCollection->addFieldToFilter('entity_id', ['in' => $this->ids]);
        }

        $products = $productCollection->getItems();

        $defaultStore = $this->storeManager->getDefaultStoreView();

        $this->emulation->startEnvironmentEmulation(
            $defaultStore ? $defaultStore->getId() : 0,
            Area::AREA_FRONTEND,
            true
        );

        try {

            /** @var ProductInterface|Product $product */
            foreach ($products as $product) {
                $entityId = (int)$product->getId();
                $this->items[$entityId] = $this->getRelatedProduct($product);
            }
        } catch (Throwable $throwable) {
            $this->logger->error($throwable->getMessage(), ['exception' => $throwable]);
        } finally {
            $this->emulation->stopEnvironmentEmulation();
        }

        if (!$this->hasReadHandler()) {
            return;
        }

        // TODO: Handle pagination
        // Temporary workaround for Morato
        $pageSize = $this->getBulkSize();
        $size = count($this->items);

        while ($size > $pageSize) {
            $data = array_slice($this->items, 0, $pageSize);
            $this->getReadHandler()->onRead($data, $pageSize, $this->getReadHandlerArguments());
            $size -= $pageSize;
        }

        if ($size > 0) {
            $this->getReadHandler()->onRead($this->items, $size, $this->getReadHandlerArguments());
        }
    }

    /**
     * @throws LocalizedException
     */
    public function getRelatedProduct(ProductInterface $product)
    {
        return new ProductRelated([
            'id'       => (int)$product->getId(),
            'upsell'   => $this->getUpSellProductIds($product),
            'related'  => $this->getRelatedProductsIds($product),
            'crossell' => $this->getCrossSellProductIds($product)
        ]);
    }

    /**
     * @param Product|ProductInterface $product
     * @return array
     * @throws LocalizedException
     */
    public function getUpSellProductIds($product)
    {
        $ids = [];

        switch ($this->getUpsellBehaviour()) {
            case Rule::BOTH_SELECTED_AND_RULE_BASED:
                $ids = array_merge(
                    $product->getUpSellProductIds(),
                    $this->upsellFactory->create()->setProduct($product)->getAllIds()
                );
                break;
            case Rule::SELECTED_ONLY:
                $ids = $product->getUpSellProductIds();
                break;
            case Rule::RULE_BASED_ONLY:
                $ids = $this->upsellFactory->create()->setProduct($product)->getAllIds();
                break;
        }

        return array_map('intval', $ids);
    }

    /**
     * @return int
     * @throws LocalizedException
     */
    public function getUpsellBehaviour()
    {
        if (!$this->upsellBehaviour) {
            $this->upsellBehaviour = (int)$this->targetHelper->getShowProducts(Rule::UP_SELLS);
        }
        return $this->upsellBehaviour;
    }

    /**
     * @param Product|ProductInterface $product
     * @return array
     * @throws LocalizedException
     */
    public function getRelatedProductsIds($product)
    {
        $ids = [];

        switch ($this->getRelatedBehaviour()) {
            case Rule::BOTH_SELECTED_AND_RULE_BASED:
                $ids = array_merge(
                    $product->getRelatedProductIds(),
                    $this->relatedFactory->create()->setProduct($product)->getAllIds()
                );
                break;
            case Rule::SELECTED_ONLY:
                $ids = $product->getRelatedProductIds();
                break;
            case Rule::RULE_BASED_ONLY:
                $ids = $this->relatedFactory->create()->setProduct($product)->getAllIds();
                break;
        }

        return array_map('intval', $ids);
    }

    /**
     * @return int
     * @throws LocalizedException
     */
    public function getRelatedBehaviour()
    {
        if (!$this->relatedBehaviour) {
            $this->relatedBehaviour = (int)$this->targetHelper->getShowProducts(Rule::RELATED_PRODUCTS);
        }
        return $this->relatedBehaviour;
    }

    /**
     * @param Product|ProductInterface $product
     * @return array
     */
    public function getCrossSellProductIds($product)
    {
        $ids = [];

        switch ($this->getCrosssellBehaviour()) {
            case Rule::BOTH_SELECTED_AND_RULE_BASED:
                $ids = array_merge(
                    $product->getCrossSellProductIds(),
                    array_map(static function($a) {return (int)$a->getId();}, $this->crosssellFactory->create()->setProduct($product)->getItemCollection())
                );
                break;
            case Rule::SELECTED_ONLY:
                $ids = $product->getCrossSellProductIds();
                break;
            case Rule::RULE_BASED_ONLY:
                $ids = $this->crosssellFactory->create()->setProduct($product)->getItemCollection();
                break;
        }

        return array_map('intval', $ids);
    }

    /**
     * @return int
     * @throws LocalizedException
     */
    public function getCrosssellBehaviour()
    {
        if (!$this->crosssellBehaviour) {
            $this->crosssellBehaviour = (int)$this->targetHelper->getShowProducts(Rule::CROSS_SELLS);
        }
        return $this->crosssellBehaviour;
    }
}
